//
//  peerguardnf (PeerGuard for Linux, BSD, and MacOS X systems)
//  Copyright (C) 2005 bitrot
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Id: BlockList.cpp,v 1.2 2005/03/27 01:25:06 nullcontext Exp $

#include "BlockList.h"
#include "PeerGuard.h"

#include <list>
#include <iostream>
#include <fstream>

using namespace std;

namespace PG {

BlockList::BlockList()
   : block_cnt_(0)
{
   pthread_mutex_init(&lock_, 0);
}

BlockList::~BlockList()
{
   pthread_mutex_destroy(&lock_);
}

// NOTE: addr should be in network byte order
bool BlockList::Find(Range* found, uint32_t addr)
{
   // First check active connections
   if (connections_.find(addr) != connections_.end())
      return false;

   BlockTable::const_iterator find;
   if ((find = binary_search_val(block_table_.begin(), block_table_.end(), IPAddr(ntohl(addr)))) != block_table_.end())
   {
      if (found)
         *found = *find;
      return true;
   }
   else
   {
      // FIXME: these never get removed...
      connections_.insert(addr);
      return false;
   }
}

int BlockList::ReadFile(const string& filename)
{
   if (filename == "-")
      return ReadStream(cin);
   else
   {
      ifstream infile(filename.c_str());
      if (!infile.is_open())
         return -1;

      return ReadStream(infile);
   }
}

int BlockList::SaveFile(const string& filename)
{
   Lock();
   FILE* fp = fopen(filename.c_str(), "wb");
   
   BlockTable::iterator itr = block_table_.begin();
   for (; itr != block_table_.end(); itr++)
   {
      uint32_t sa0 = ntohl(itr->start);
      uint32_t ea0 = ntohl(itr->end);
      char* sa = (char*)&sa0;
      char* ea = (char*)&ea0;
      fprintf(fp, "%s:%d.%d.%d.%d-%d.%d.%d.%d\n", itr->rule.c_str(),
         sa[0] & 0xFF, sa[1] & 0xFF, sa[2] & 0xFF, sa[3] & 0xFF,
         ea[0] & 0xFF, ea[1] & 0xFF, ea[2] & 0xFF, ea[3] & 0xFF);
   }

   fclose(fp);
   Unlock();

   return 0;
}

int BlockList::AddRange(const string& rulename, uint32_t start, uint32_t end)
{
   Lock();
   connections_.clear();
   block_table_.push_back(Range(start, end, rulename));
   Unlock();
   Merge();
   return 0;
}

// FIXME: there could be multiple ranges with the same name
int BlockList::RemoveRange(const string& rulename)
{
   Lock();
   BlockTable::iterator itr = block_table_.begin();
   while (itr != block_table_.end())
   {
      if (itr->rule == rulename)
         block_table_.erase(itr++);
      else
         itr++;
   }

   Unlock();

   return 0;
}

int BlockList::RemoveIp(uint32_t addr)
{
   Lock();
   // FIXME: this will need to split ranges
   Unlock();
   return 0;
}

int BlockList::Merge()
{
   Lock();
   sort(block_table_.begin(), block_table_.end());

   list<Range> temp_list;

   // convert iptable to a list so we can quickly erase elements
   {
      BlockTable::iterator itr = block_table_.begin();
      for (; itr != block_table_.end(); itr++)
         temp_list.push_back(*itr);
   }

   list<Range>::iterator itr1 = temp_list.begin();
   for (; itr1 != temp_list.end(); itr1++)
   {
      list<Range>::iterator itr2 = itr1;
      itr2++;
      while (itr2 != temp_list.end() && *itr1 == *itr2)
      {
         uint32_t nstart = itr1->start < itr2->start ? itr1->start : itr2->start;
         uint32_t nend = itr1->end > itr2->end ? itr1->end : itr2->end;
         itr1->start = nstart;
         itr1->end = nend;
         temp_list.erase(itr2++);
      }
   }

   // convert the list back to iptable vector format
   {
      block_table_.clear();
      list<Range>::iterator itr = temp_list.begin();
      for (; itr != temp_list.end(); itr++)
         block_table_.push_back(*itr);
   }

   Unlock();

   return 0;
}

int BlockList::ReadStream(istream& in)
{
   Lock();

   int binary_mode = 0;
   char delim = '\n';
   int temp = in.peek();
   if (isascii(temp))
      *gLogStream << "detected ASCII blocklist" << endl;
   else
   {
      char buf[7];
      int version;
      in.read(buf, sizeof(buf));
      version = (int)in.get();
      if (!memcmp(buf, "\xFF\xFF\xFF\xFFP2B", 7))
      {
         *gLogStream << "detected binary blocklist" << endl;
         binary_mode = 1;
         delim = 0x00;
	 if(version!=1 && version!=2)
			*gLogStream << "unknown p2b version";
      }
   }

   char buf[16384];
   unsigned int ipcnt = 0;

   connections_.clear();

   while (in.getline(buf, sizeof(buf), delim))
   {
      string rule;
      char* pos;
      uint32_t addr_1, addr_2;

      if (binary_mode)
      {
         rule = buf;
         in.read((char*)&addr_1, 4);
         in.read((char*)&addr_2, 4);
      }
      else
      {
         pos = strchr(buf, ':');
         if (!pos)
            continue;
         else
         {
            char* new_pos = pos;
            do
            {
               pos = new_pos;
               new_pos = strchr(pos + 1, ':');
            } while(new_pos);
         }
   
         rule.assign(buf, pos-buf);
   
         pos++;
         int ip1_1, ip1_2, ip1_3, ip1_4, ip2_1, ip2_2, ip2_3, ip2_4;
         int count = sscanf(pos, "%d.%d.%d.%d-%d.%d.%d.%d",
            &ip1_1, &ip1_2, &ip1_3, &ip1_4, &ip2_1, &ip2_2, &ip2_3, &ip2_4);
   
         if (count < 8)
         {
            *gLogStream << "failed to read IP addresses: " << buf << endl;
            continue;
         }

         pos = (char*)&addr_1;
         pos[0] = (ip1_1 & 0xFF);
         pos[1] = (ip1_2 & 0xFF);
         pos[2] = (ip1_3 & 0xFF);
         pos[3] = (ip1_4 & 0xFF);
         pos = (char*)&addr_2;
         pos[0] = (ip2_1 & 0xFF);
         pos[1] = (ip2_2 & 0xFF);
         pos[2] = (ip2_3 & 0xFF);
         pos[3] = (ip2_4 & 0xFF);
      }

      addr_1 = ntohl(addr_1);
      addr_2 = ntohl(addr_2);

      block_table_.push_back(Range(addr_1, addr_2, rule));

      ipcnt += addr_2 - addr_1 + 1;
   }

   Unlock();

   Merge();

   return 0;
}

} // namespace PG

