//
//  peerguardnf (PeerGuard for Linux, BSD, and MacOS X systems)
//  Copyright (C) 2005 bitrot
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Id: BlockList.h,v 1.2 2005/03/27 01:25:14 nullcontext Exp $

#ifndef BLOCKLIST_HEADER___
#define BLOCKLIST_HEADER___

#include <pthread.h>
#include <istream>
#include <netinet/in.h>
#include <string>
#include <vector>
#include <ext/hash_set>
#include <sstream>

namespace PG {

struct Range;

///////////////////////////////////////////////////////////////////////////////
// An IP address.  This allows sorting and checking if the address falls inside
// a Range.
///////////////////////////////////////////////////////////////////////////////
struct IPAddr
{
   uint32_t ip;

   IPAddr() : ip(0) {}
   IPAddr(uint32_t addr) : ip(addr) {}
   operator uint32_t() const { return ip; }

   // sorting operator
   inline bool operator <(const Range& r) const;

   // tests if this IP is inside a Range
   inline bool operator ==(const Range& r) const;
   
   std::string ToString() const
   {
      std::ostringstream out;
      uint32_t t_ip = htonl(ip);
      char* addr = (char*)&t_ip;
      out << (unsigned int)(addr[0] & 0xFF) << "." 
          << (unsigned int)(addr[1] & 0xFF) << "." 
          << (unsigned int)(addr[2] & 0xFF) << "." 
          << (unsigned int)(addr[3] & 0xFF);
      return out.str();
   }
};

///////////////////////////////////////////////////////////////////////////////
// An IP address range.  This allows sorting and other test functionality.
///////////////////////////////////////////////////////////////////////////////
struct Range
{
   IPAddr start;
   IPAddr end;
   std::string rule;

   Range() : start(0), end(0) {}
   Range(uint32_t s, uint32_t e, const std::string& r) : start(s), end(e), rule(r) {}
   bool operator ==(const IPAddr& addr) const { return (addr.ip >= start.ip && addr.ip <= end.ip); }
   bool operator <(const IPAddr& addr) const { return end < addr; }
   bool operator <(const Range& r) const { return start < r.start; }

   bool operator ==(const Range& r) const
   {
      // Test intersection of two ranges
      return (start == r || end == r || r.start == *this || r.end == *this);
   }
   
   std::string ToString() const
   {
      std::ostringstream out;
      out << rule << ":" << start.ToString() << "-" << end.ToString();
      return out.str();
   }
};

inline bool IPAddr::operator <(const Range& r) const { return ip < r.start; }
inline bool IPAddr::operator ==(const Range& r) const { return (ip >= r.start.ip && ip <= r.end.ip); }

typedef std::vector<Range> BlockTable;

///////////////////////////////////////////////////////////////////////////////
// BlockList
///////////////////////////////////////////////////////////////////////////////
class BlockList
{   

   pthread_mutex_t lock_;
   BlockTable block_table_;
   size_t block_cnt_;
   __gnu_cxx::hash_set<uint32_t> connections_;

public:
   BlockList();
   ~BlockList();

   bool Find(Range* found, uint32_t addr);
   int ReadFile(const std::string& filename);
   int SaveFile(const std::string& filename);
   int AddRange(const std::string& rulename, uint32_t start, uint32_t end);
   int RemoveRange(const std::string& rulename);
   int RemoveIp(uint32_t addr);

   const BlockTable& Table() const { return block_table_; }

   size_t IpCount() const
   {
      size_t ipcnt = 0;
      BlockTable::const_iterator itr = block_table_.begin();
      for (; itr != block_table_.end(); itr++)
         ipcnt += itr->end - itr->start + 1;
      return ipcnt;
   }

   size_t RangeCount() const { return block_table_.size(); }
   size_t BlockCount() const { return block_cnt_; }
   size_t operator ++(int) { block_cnt_++; return block_cnt_; }

   void Lock() { pthread_mutex_lock(&lock_); }
   void Unlock() { pthread_mutex_unlock(&lock_); }

protected:
   int Merge();
   int ReadStream(std::istream& in);

   // STL binary_search that returns the found item
   static BlockTable::const_iterator
   binary_search_val(BlockTable::const_iterator itr1, BlockTable::const_iterator itr2, const IPAddr& addr)
   {
      BlockTable::const_iterator find_itr = lower_bound(itr1, itr2, addr);
      if (find_itr != itr2 && !(addr < *find_itr))
         return find_itr;
      else
         return itr2;
   }
};

} // namespace PG

#endif // BLOCKLIST_HEADER___

