//
//  peerguardnf (PeerGuard for Linux, BSD, and MacOS X systems)
//  Copyright (C) 2005 BitRot
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Id: HttpServer.cpp,v 1.3 2005/03/27 01:27:41 nullcontext Exp $

#include "HttpServer.h"
#include "PeerGuard.h"

#if defined(__APPLE__) || defined(__BSD__)
#include <netinet/in_systm.h>
#endif

#include <errno.h>
#include <sys/socket.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>

using namespace std;
using namespace __gnu_cxx;

namespace PG {
   
bool gStopHttpThread = true;
static pthread_t gHttpThreadHandle;
static int gSock = -1;
static int gClientSock = -1;

int WriteData(int sock, const string& data)
{
   int result = -1;
   
   while (!gStopHttpThread)
   {
      ssize_t size = write(sock, &data[0], data.size());
      if ((size < 0 && errno != EAGAIN) || (size_t)size < data.size())
         break;
      else
      {
         result = 0;
         break;
      }
   }

   return result;
}

string ReadData(int sock)
{
   string buf;
   string out;

   buf.resize(0xFFFF);

   while (!gStopHttpThread)
   {
      ssize_t size = read(sock, &buf[0], buf.size());
      if (size < 0)
      {
         if (errno != ECONNABORTED && errno != EINTR)
            *gLogStream << "Socket read() err: " << size << " " << errno << endl;
         break;
      }
      
      out.append(buf.data(), size);

      if (out.find("\r\n\r\n") != string::npos)
         break;
   }

   return out;
}

void GetURLData(CGIData* data, const string& req_)
{
   string url, options, command, req = req_;
   string::size_type start, end;

   data->clear();

   start = req.find("Cookie: PGAUTH=");
   if (start != string::npos)
   {
      string cookie = req.substr(start + strlen("Cookie: PGAUTH="));
      end = cookie.find("\r");
      if (end != string::npos)
         cookie.resize(end);

      end = cookie.find("\n");
      if (end != string::npos)
         cookie.resize(end);

      end = cookie.find(";");
      if (end != string::npos)
         cookie.resize(end);

      (*data)["PGAUTH"] = cookie;
   }

   end = req.find('\r');
   if (end != string::npos)
      req.resize(end);

   start = req.find("GET ");

   if (start != string::npos)
   {
      start += 4;
      end = req.find(" HTTP/", start);
      if (end != string::npos)
         url = string(req, start, end - start);
   }

   end = url.find('?');
   if (end != string::npos)
   {
      options = string(url, end + 1);
   }

   url = url.substr(0, end);
   while (url[url.size() - 1] == '/')
      url.resize(url.size() - 1);

   start = url.find_last_of('/');

   if (start == string::npos)
      command = url;
   else
      command = url.substr(start + 1);

   start = 0;
   do
   {
      end = options.find('&', start);
      string option = options.substr(start, end - start);
      string::size_type idx = option.find('=');
      string key, value;
      if (idx == string::npos)
         key = option;
      else
      {
         key = option.substr(0, idx);
         value = option.substr(idx + 1);
      }

      (*data)[key] = value;
         
      start = end + 1;
   } while (end != string::npos);

   (*data)["URL"] = url;
   (*data)["COMMAND"] = command;
   (*data)["OPTIONS"] = options;
}

void* HttpThread(void* block_list_)
{
   BlockList* block_list = (BlockList*)block_list_;
   
   gSock = socket(AF_INET, SOCK_STREAM, 0);
   if (gSock < 0)
   {
      *gLogStream << "HTTP server could not create socket" << endl;
      return 0;
   }
   
   struct sockaddr_in server;
   memset(&server, 0, sizeof(server));

   server.sin_family = AF_INET;
   server.sin_port = htons(5555);
   server.sin_addr.s_addr = INADDR_ANY;

   if (bind(gSock, (sockaddr*)&server, sizeof(server)) < 0)
   {
      *gLogStream << "HTTP server bind error" << endl;
      close(gSock);
      exit(1);
   }

   if (listen(gSock, 5) < 0)
   {
      *gLogStream << "HTTP server listen error" << endl;
      close(gSock);
      exit(1);
   }
   
   struct sockaddr_in client;
   memset(&client, 0, sizeof(client));

   socklen_t clientlen = sizeof(client);
   string buf;

   int bTrue = 1;
   setsockopt(gSock, SOL_SOCKET, SO_LINGER, &bTrue, sizeof(bTrue));
   
   // Accept connections
   while (!gStopHttpThread)
   {
      gClientSock = accept(gSock, (sockaddr*)&client, &clientlen);     
      if (gClientSock < 0)
      {
         if (errno != ECONNABORTED && errno != EINTR)
            *gLogStream << "Socket accept() err: " << gClientSock << " " << errno << endl;

         break;
      }
      
      buf = ReadData(gClientSock);
         
      HandleHttpRequest(gClientSock, buf, block_list);
      
      if (gClientSock != -1)
      {
         close(gClientSock);
         gClientSock = -1;
      }
   }
   
   if (gSock != -1)
      close(gSock);

   return 0;
}

int StartHttpServer(BlockList* blocklist)
{
   if (!gStopHttpThread)
      return -1;
   else
   {
      gStopHttpThread = false;
      return pthread_create(&gHttpThreadHandle, 0, HttpThread, blocklist);
   }
}

int StopHttpServer()
{
   if (gStopHttpThread)
      return -1;
   else
   {
      if (gSock != -1)
         close(gSock);

      if (gClientSock != -1)
         close(gClientSock);

      gStopHttpThread = true;
      void* result;
      pthread_join(gHttpThreadHandle, &result);
      return 0;
   }
}

} // namespace PG

