//
//  peerguardnf (PeerGuard for Linux, BSD, and MacOS X systems)
//  Copyright (C) 2005 bitrot
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Id: PeerGuard.cpp,v 1.2 2005/03/27 01:29:25 nullcontext Exp $

#include "PeerGuard.h"
#include "BlockList.h"

#include <sys/wait.h>
#include <fstream>

using namespace std;

namespace PG {

ostream* gLogStream = 0;
PGPreferences gPrefs;
bool gStopBlockerThread = true;

static pthread_t gBlockerThreadHandle;

int PGPreferences::Read(const string& filename)
{
   ifstream file(filename.c_str());
   if (!file.is_open())
      return -1;

   char buf[16384];

   while (file.getline(buf, sizeof(buf)))
   {
      char* pos = strchr(buf, '#');
      if (pos)
         *pos = 0;

      char* temp;
      char* param;
      pos = strtok_r(&buf[0], " =", &temp);
      if (pos)
      {
         char* temp;
         param = &buf[strlen(pos) + 1];

         if (!strcasecmp(pos, "BLOCKFILE"))
         {
            char* filename = strtok_r(param, " =,;", &temp);

            if (filename)
               gPrefs.blocklistFilename = filename;
         }
         else if (!strcasecmp(pos, "PORTS_NOT_TO_BLOCK"))
         {
            char* port = strtok_r(param, " =,;", &temp);
            while (port)
            {
               gPrefs.AllowPort(atoi(port));
               port = strtok_r(0, " ,;", &temp);
            }
         }
         else if (!strcasecmp(pos, "ADMIN_NAME"))
         {
            char* name = strtok_r(param, " =,;", &temp);

            if (name)
               gPrefs.adminName = name;
         }
         else if (!strcasecmp(pos, "ADMIN_PASSWORD"))
         {
            char* password = strtok_r(param, " =,;", &temp);

            if (password)
               gPrefs.adminPassword = password;
         }
      }
   }

   return 0;
}

//
// Executes a system command similar to system() except safe.  Note that the
// functionality is extremely basic.  Arguments are separated by spaces and
// no quoting is possible.
//
int SafeExec(const char* cmd, bool output = true)
{
   int status = 0;

   pid_t pid = fork();
   if (pid == -1)
      return -1;
   else if (!pid)
   {
      vector<char*> args;
      string cmdline = cmd;

      cmdline += '\000';

      char* temp;
      char* pos = strtok_r(&cmdline[0], " ", &temp);
      while (pos)
      {
         args.push_back(pos);
         pos = strtok_r(0, " ", &temp);
      }

      args.push_back(pos);

      if (!output)
      {
         close(0);
         close(2);
      }

      execvp(args[0], &args[0]);
   }
   else
      waitpid(pid, &status, 0);

   return status;
}

int StartBlockerThread(BlockList* blocklist)
{
   if (!gStopBlockerThread)
      return -1;
   else
   {
      gStopBlockerThread = false;
      return pthread_create(&gBlockerThreadHandle, 0, BlockerThread, blocklist);
   }
}

int StopBlockerThread()
{
   if (gStopBlockerThread)
      return -1;
   else
   {
      gStopBlockerThread = true;
      void* result;
      pthread_join(gBlockerThreadHandle, &result);
      return 0;
   }
}

int CreateFirewallRules()
{
#if defined(__APPLE__) || defined(__BSD__)
   // This is allowed to fail as the rule probably does not exist
   SafeExec("ipfw delete 02045", false);

   if (SafeExec("ipfw add 02045 divert 6137 all from any to any"))
   {
      *gLogStream << "Failed to set rules with ipfw" << endl;
      return -1;
   }

#else // Linux
   // These are allowed to fail as the rule probably does not exist
   SafeExec("iptables -D INPUT -j QUEUE", false);
   SafeExec("iptables -D OUTPUT -j QUEUE", false);

   if (SafeExec("iptables -I INPUT -j QUEUE"))
   {
      *gLogStream << "Failed to set INPUT rule with iptables" << endl;
      return -1;
   }

   if (SafeExec("iptables -I OUTPUT -j QUEUE"))
   {
      SafeExec("iptables -D INPUT -j QUEUE", false);
      *gLogStream << "Failed to set OUTPUT rule with iptables" << endl;
      return -1;
   }
#endif

   return 0;
}

int DeleteFirewallRules()
{
#if defined(__APPLE__) || defined(__BSD__)
   SafeExec("ipfw delete 02045", false);
#else // Linux
   SafeExec("iptables -D INPUT -j QUEUE", false);
   SafeExec("iptables -D OUTPUT -j QUEUE", false);
#endif

   return 0;
}

} // namespace PG

